# Copyright (c) 2020 Stefan Neumann and Pauli Miettinen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy
import random

import centermodule
import HelperFunctions
import MisraGries

class OnlineFL:
	def __init__(self, f, maxNumCenters, maxSolutionCost):
		self.f = f
		self.maxNumCenters = maxNumCenters
		self.maxSolutionCost = maxSolutionCost

		self.objectiveFunction = 0
		self.centers = []
		self.MGs = []

	''' Returns True if p was successfully added and False if there are too many
		open centers or the objective function would be too high after adding
		the point.
	'''
	def addPoint(self, p, MG):
		dist, closestCenterIndex = centermodule.closestCenter(self.centers, p, HelperFunctions.asymHammingDistanceWeight)
		'''
		# IF YOU DO NOT WANT TO COMPILE CENTERMODULE, YOU CAN USE
		# THE FOLLOWING PYTHON CODE, BUT THIS WILL BE MUCH SLOWER

		dist, closestCenterIndex = HelperFunctions.distanceToClosestCenter(self.centers, p)
		'''

		weight = MG.numVertices
		probToOpenCenter = weight * dist / self.f
		if probToOpenCenter >= 1:
			shouldOpenCenter = True
		else:
			shouldOpenCenter = random.random() < probToOpenCenter

		if self.triggersStoppingCriterion(shouldOpenCenter, dist, weight):
			return False

		if shouldOpenCenter:
			self.openNewCenter(p, MG)
		else:
			self.addPointToCenter(p, dist, closestCenterIndex, MG)

		return True

	def addPointsFromCenters(self, points, MGs):
		for i in range( len(points) ):
			self.addPoint(points[i], MGs[i])

	def triggersStoppingCriterion(self, shouldOpenCenter, dist, weight):
		if shouldOpenCenter: # we would open too many centers
			return (len(self.centers) >= self.maxNumCenters)
		else: # no new center is opened but the objective function might become too large
			newObjectiveFunction = self.objectiveFunction + weight*dist
			return (newObjectiveFunction > self.maxSolutionCost)

	def openNewCenter(self, p, MG):
		self.centers.append(p)
		self.MGs.append(MG)

	def addPointToCenter(self, p, dist, centerIndex, MG):
		weight = MG.numVertices
		self.objectiveFunction += weight * dist

		self.MGs[centerIndex].addCounters(MG)
		self.MGs[centerIndex].removeSuperfluousCounters()

