#!/usr/bin/env python3

'''
Please see the README file for instructions of how to run sofa.
Use option -h to see an exhaustive list of all options.
'''
# Copyright (c) 2020 Stefan Neumann and Pauli Miettinen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy
import scipy
import random
import sys, getopt
import os

import HelperFunctions

import Sofa

def main(argv):
	' I/O options '
	inputfile = ''
	outputfile = ''
	writeClusterOutput = False
	leftClusterOutputFile = ''
	sparseOutput = False
	fileIsSparse=False
	skipHeader=False

	' options for sofa '
	k=20
	thresholds = []
	verbose = 0
	clustering = 'doNothing' # or 'kmeans'
	minClusterSize = 20

	maxNumCenters=200
	beta=2
	MGMaxCounters=2000
	s=500
	asymHammingDistanceWeight=0.1

	coverLeftSide=None
	minOverlap=0.0 	
	try:
		longOptions = ["ifile=",
					   "ofile=",
					   "threshold=",
					   "minClusterSize=",
					   "clustering=",
					   "verbose=",
					   "leftClusterOutputFile=",
					   "maxNumCenters=",
					   "beta=",
					   "MGMaxCounters=",
					   "minClusterSize=",
					   "s=",
					   "asymHammingDistanceWeight=",
					   "coverLeftSide=",
					   "minOverlap=",
					   "fileIsSparse",
					   "sparseOutput",
					   "help",
					   "skipHeader"]
		opts, args = getopt.getopt(argv,"hk:i:o:",longOptions)
	except getopt.GetoptError as err:
		print(f'Got the following error when reading input arguments: {err}')
		print('Use option -h to see a list of all available options.')
		sys.exit(2)
	for opt, arg in opts:
		if opt in ("-i", "--ifile"):
			inputfile = arg
		elif opt in ("-o", "--ofile"):
			outputfile = arg
		elif opt in ("-h", "--help"):
			printHelp()
			return
		elif opt == "--leftClusterOutputFile":
			writeClusterOutput = True
			leftClusterOutputFile = arg
		elif opt == "--maxNumCenters":
			maxNumCenters = int(arg)
		elif opt == "--beta":
			beta = float(arg)
		elif opt == "--MGMaxCounters":
			MGMaxCounters = int(arg)
		elif opt == "--s":
			s = int(arg)
		elif opt == "--asymHammingDistanceWeight":
			asymHammingDistanceWeight = float(arg)
		elif opt == "--coverLeftSide":
			coverLeftSide = arg
		elif opt == "--minOverlap":
			minOverlap = float(arg)
		elif opt == "--fileIsSparse":
			fileIsSparse = True
		elif opt == "--sparseOutput":
			sparseOutput = True
		elif opt == "--skipHeader":
			skipHeader = True
		elif opt == "-k":
			k = int(arg)
		elif opt == "--threshold":
			threshold = arg
			if arg != 'auto':
				try:
					threshold = float(arg)
				except ValueError:
					print(f'Cannot parse threshold {threshold}.')
			thresholds.append(threshold)
		elif opt == "--minClusterSize":
			minClusterSize = int(arg)
		elif opt == "--verbose":
			verbose = int(arg)
		elif opt == "--clustering":
			if arg == "kmeans" or "doNothing":
				clustering = arg
			else:
				print('Unknown clustering method given.')
				sys.exit()
		else:
			print("unknown option {0}".format(opt))
			sys.exit()

	if len(thresholds) == 0:
		thresholds.append('auto')

	if len(thresholds) == 0:
		thresholds = ['auto']

	rightClusters,leftClusters,rows,columns = Sofa.sofa(
			inputfile,
			k,
			thresholds=thresholds,
			maxNumCenters=maxNumCenters,
			beta=beta,
			MGMaxCounters=MGMaxCounters,
			minClusterSize=minClusterSize,
			clusteringMethod=clustering,
			s=s,
			asymHammingDistanceWeight=asymHammingDistanceWeight,
			fileIsSparse=fileIsSparse,
			skipHeader=skipHeader,
			coverLeftSide=coverLeftSide
	)

	' write the output to disk '
	for threshold in thresholds:
		clustersfile = outputfile
		leftclustersfile = leftClusterOutputFile
		if len(thresholds) > 1:
			clustersfile += '_' + str(threshold)
			leftclustersfile += '_' + str(threshold)

		writeRightClustersForFixedThreshold(clustersfile,
											rightClusters[threshold],
											sparseOutput,
											columns=columns)
		if writeClusterOutput:
			if coverLeftSide == None:
				print('Sofa was run with coverLeftSide=None and therefore did not compute '
						' left clusters, but now it is asked to write the left '
						' clusters to a file (option --leftClusterOuputFile was '
						' set). Not writing any output file.')
				return

			writeLeftClustersForFixedThreshold(leftclustersfile,
											   leftClusters[threshold],
											   sparseOutput,
											   writeEmptyClusters=True,
											   rows=rows)

	if verbose > 0:
		printResultsToScreenForThresholds(rightClusters, leftClusters, thresholds, verbose)


def writeRightClustersForFixedThreshold(outputfile,
										rightClusters,
										sparseOutput,
										columns=0):
	if sparseOutput:
		HelperFunctions.writeReconstructedClustersToFile(outputfile, rightClusters)
	else:
		HelperFunctions.writeClusterMatrix(outputfile, rightClusters, columns)

def writeLeftClustersForFixedThreshold(leftClusterOutputFile,
									   leftClusters,
									   sparseOutput,
									   writeEmptyClusters=True,
									   rows=0):
	if sparseOutput:
		HelperFunctions.writeReconstructedClustersToFile(leftClusterOutputFile, leftClusters, writeEmptyClusters=True)
	else:
		HelperFunctions.writeClusterMatrix(leftClusterOutputFile, leftClusters, rows)

def printResultsToScreenForThresholds(rightClusters, leftClusters, thresholds, verbose):
	for threshold in thresholds:
		print(f'\nResults for threshold {threshold}:')
		printResultsToScreen(rightClusters[threshold],
							 leftClusters[threshold],
							 verbose)

def printResultsToScreen(rightClusters, leftClusters, verbose):
	if verbose > 0:
		print('\t- number of rightClusters: {0}. with lengths: '.format(len(rightClusters)), end='')
		for cl in rightClusters:
			print('{0}, '.format(len(cl)), end='')
		print('\b\b ')

	if verbose >= 2:
		print('\t- number of leftClusters: {0}. with lengths: '.format(len(leftClusters)), end='')
		for cl in leftClusters:
			print('{0}, '.format(len(cl)), end='')
		print('\b\b ')

	if verbose == 3:
		print('\trightClusters = {0}'.format(rightClusters))
		print('\tleftClusters = ',end='')
		for cl in leftClusters:
			if len(cl) < 20 and verbose != 4:
				continue
			print('\n- {0}, '.format(sorted(cl)),end='')
		print('\b\b ')

def printHelp():
	print('The sofa algorithm finds clusters in bipartite graphs in a streaming\n'
			'setting. The implementation is based on the following publication:\n'
			'	Biclustering and Boolean Matrix Factorization in Data Streams\n'
			'	by Stefan Neumann and Pauli Miettinen\n'
			'	PVLDB, 13(10): 1709-1722, 2020\n'
			'	DOI: https://doi.org/10.14778/3401960.3401968\n'
			'See README for how to compile the centermodule, sofa\'s requirements\n'
			'and the license of the code.\n')
	print('LIST OF OPTIONS')
	print('---------------')
	print('-i X')
	print('	Sets the input file to X. Assumes that X is a binary biadjacency matrix\n'
		   	'	of a bipartite graph stored as a csv file,\n'
			'	where each row corresponds to a row of the matrix (=a left-side vertex of the graph)\n'
			'	and entries are separated by commas.\n'
			'	The file will be read line by line and the lines correspond\n'
			'	to the data stream.\n'
			'	If the input file is sparse, set the flag --fileIsSparse.')
	print('--ifile X')
	print('	Equivalent to using "-i X".')
	print('--fileIsSparse')
	print('	When this option is passed, then the input file is expected to be in\n'
			'	the following csv-format where each line looks like this:\n'
			'		leftVtxId,rightVtxId,...\n'
			'	This corresponds to entry (leftVtxId, rightVtxId) of the\n'
			'	biadjacency matrix being 1; the "..." part can be arbitrary\n'
			'	and will be ignored by sofa.\n'
			'	Assumes that all lines starting with leftVtxId appear consecutively\n'
			'	(i.e., for each left vertex all of its right-side neighbors\n'
			'	appear consecutively).')
	print('--skipHeader')
	print('	If this option is passed, sofa will skip the first line of the\n'
			'	input file.')
	print('-o X')
	print('	Writes the right clusters to X.\n'
			'	By default, the output file contains a dense factor matrix.\n'
			'	The dense factor matrices are stored in a csv format, where\n'
			'	each row corresponds to a cluster and the non-zero entries\n'
			'	indicate the vertices which are part of the cluster.\n'
			'	For example, if a line reads "0,1,1,0,0,0,1" (indeed, the real\n'
			'	files contain floating point values instead of integers),\n'
			'	then vertices 1, 2 and 6 are contained in the cluster.\n'
			'	For writing the left clusters, use option "--leftClusterOuputFile X".\n'
			'	To obtain a sparse output, use option "--sparseOutput".')
	print("--ofile X")
	print('	Equivalent to using "-o X".')
	print('--leftClusterOutputFile X')
	print('	Writes the left clusters to file X.\n'
			'	To use this option, you also have to set the flag "--coverLeftSide X".\n'
			'	See option "-o X" for information on the output format.\n'
			'	Note that the cluster in the i\'th row of the *left* cluster output file\n'
			'	belongs to the i\'th cluster in the *right* cluster output file.')
	print('--sparseOutput')
	print('	When this flag is passed, the output will be written in a sparse format.\n'
			'	In this case, each line of the output file corresponds to a cluster\n'
			'	and lists the IDs of vertices that are contained in the cluster.\n'
			'	The vertex IDs are separated by commas.\n'
			'	For example, if a line of the file is "5,12,1909,20000",\n'
			'	then vertices 5, 12, 1909 and 20000 are part of that cluster.')
	print('--coverLeftSide X')
	print('	When this option is used, sofa also computes the clusters of the\n'
			'	left-side vertices.\n'
			'	Use X="FindAllAssoStyleTopK" to get BMF-style covering (Section 4.2).\n'
			'	Use X="FindOne" to get biclustering-style covering (Section 4.1).')
	print('--minOverlap X')
	print('	When minOverlap is set a value 0<X<1, then\n'
			'	the biclustering-style covering only covers a left-side\n'
			'	vertex if the fraction in Equation (2) in Section 4.1 of the paper\n'
			'	evaluates to at least X. By default, X=0.0.')
	print('-h')
	print('	Lists and explains all options of the command-line program.')
	print('--help')
	print('	Equivalent to -h.')
	print('-maxNumCenters X')
	print('	Sets the maximum number of centers that sofa is allowed to maintain\n'
			'	to X. In our real-world experiments, we used X=20*k.')
	print('--beta X')
	print('	beta is the parameter that states how the lower bound in Line 20 of\n'
			'	sofa grows. In the paper and our experiments, we simply used X=2.')
	print('--MGMaxCounters X')
	print('	Sets the number of counters used in the heavy hitters data\n'
			'	structures to X. In our real-world experiments, we used\n'
			'	X=min{3*s,0.05*n}, where s is the maximum degree of the vertices\n'
			'	and n is the number of left-side vertices of the graph.')
	print('--s X')
	print('	Sets the maximum degree of the vertices to s. All vertices with\n'
			'	degree larger than s will not be processed by sofa.')
	print('--asymHammingDistanceWeight X')
	print('	Sets the asymmetric Hamming distance weight to X.\n'
			'	See Section 5.1 of the paper. The default is X=0.1 and we also\n'
			'	used this value in our experiments.')
	print('-k X')
	print('	Sets k, the number of clusters that shall be computed, to X.')
	print('--threshold X\n'
		'	Sets the rounding threshold used in sofa\'s postprocessing to X\n'
		'	(Line 25 in the pseudocode).\n'
		'	To use the heuristic for determining the threshold (Section 5.4),\n'
		'	use X="auto".\n'
		'	It is possible to use this option multiple times. If multiple thresholds\n'
		'	are passed, multiple output files will be created, each of them containing\n'
		'	the corresponding threshold as a suffix. For example, if you pass\n'
		'	"--threshold 0.6 --threshold 0.8 --threshold auto", sofa will create\n'
		'	output files "output_0.6", "output_0.8" and "output_auto".')
	print('--minClusterSize X')
	print('	When minClusterSize is set to X, the postprocessing step of sofa\n'
			'	(Lines 21--25 in the pseudocode) are only run for centers\n'
			'	that have been assigned at least X points.\n'
			'	This is useful to remove noisy outputs.\n'
			'	Default: X=20.')
	print('--verbose X')
	print('	Determines the verbosity of sofa\'s on-screen output.\n'
			'	If X=0, sofa\'s output is not printed on screen.\n'
			'	If X>=1, sofa writes the number of right clusters and how many\n'
			'	vertices were assigned to each cluster.\n'
			'	If X>=2, sofa writes the number of left clusters and how many\n'
			'	vertices were assigned to each cluster.\n'
			'	If X>=3, sofa writes the right and left clusters it found.\n'
			'	Default: X=0.')
	print('--clustering X')
	print('	The clustering method that shall be used in the postprocessing step\n'
			'	of sofa (Line 21 in the pseudocode).\n'
			'	The options are "kmeans" and "doNothing".\n'
			'	The "kmeans" option is advisable for biclustering algorithms\n'
			'	(see Section 5.2) and the "doNothing" option is advisable\n'
			'	for BMF algorithms (see Section 5.3).')
	print('\nEXAMPLES')
	print('--------')
	print('BMF algorithm:')
	print('	python3 main.py -k 20 --maxNumCenters 200 --MGMaxCounter 2000 --s 500 --clustering doNothing --threshold auto --minClusterSize 20 --coverLeftSide FindAllAssoStyleTopK -i INPUT_FILE (--fileIsSparse) -o RIGHT_CLUSTER_OUTPUT --leftClusterOutputFile LEFT_CLUSTER_OUTPUT (--sparseOutput) --verbose 1')
	print('\nbiclustering algorithm:')
	print('	python3 main.py -k 20 --maxNumCenters 200 --MGMaxCounter 2000 --s 500 --clustering kmeans --threshold auto --minClusterSize 20 --coverLeftSide FindOne -i INPUT_FILE (--fileIsSparse) -o RIGHT_CLUSTER_OUTPUT --leftClusterOutputFile LEFT_CLUSTER_OUTPUT (--sparseOutput) --verbose 1')

if __name__ == '__main__':
	main(sys.argv[1:])

